import os
from ctypes import *
import wave
import eloquence
import socket
checked = False
import time
import logger_setup
logger_setup.setup_logging()
from cStringIO import StringIO
import asyncore, asynchat
import sound_lib.output
output = sound_lib.output.Output()
from sound_lib.stream import FileStream
import command_parser

port = 64111

class Receiver(asynchat.async_chat):
 conid = 1

 def __init__(self, server, sock, addr):
  asynchat.async_chat.__init__(self, sock)
  self.chanid = self.conid
  Receiver.conid += 1  
  self.wave = None
  self.set_terminator('\n')
  self.data = ''
  self.server = server
  self.output_io = StringIO()
  self.synth = eloquence.Synth()
  samples = 1500
  self.buf = create_string_buffer(samples*2)
  self.fs = None
  self.parser = command_parser.Parser()
  self.ctxt = 'norm'
  self.pitch = 65

  Callback = CFUNCTYPE(c_int, c_int, c_int, c_int, c_void_p)
  @Callback
  def callback(h, msg, lp, d):
   if msg == 0:
    self.wave.writeframes(string_at(self.buf, lp*2))
   return 1
  self.callback = callback
  self.synth.registerCallback(callback, 0)
  self.synth.setParam(eloquence.eciInputType, 1)
  self.synth.setParam(eloquence.eciSynthMode, 1)
  self.synth.setOutputBuffer(samples, self.buf)
  self.rate = 57
  eloquence.lib.eciSetVoiceParam(self.synth.eci, 0, eloquence.eciSpeed, self.rate)

 def done(self):
  self.push('x\n')

 def collect_incoming_data(self, data):
  self.data += data

 def found_terminator(self):
  global checked
  data = self.data.strip()
  self.data = ''
  if not data: return
  if data[0] == 's':
   self.reset_wave()
   d = data[1:].decode('utf-8').encode('mbcs', 'replace')
   d = d.strip(',')
   if not d:
    self.push('x\n')
    return
   sequence = self.parser.parse(d)
   self.speak_sequence(sequence)
   self.wave.close()
   if self.wave.getnframes() == 0:
    self.push('x\n')
    return
   outbuf = create_string_buffer(self.output_io.getvalue(), self.output_io.tell())
   self.fs = FileStream(mem=True, file=outbuf, length=-1)
   self.fs.play()
  elif data[0] == 'x':
   self.synth.stop()
   self.stop()
  elif data[0] == 'r':
   rate = int(data[1:])
   rate = self._paramToPercent(rate, 0, 500)
   rate = self._percentToParam(rate, 40, 150)
   self.rate = rate
   eloquence.lib.eciSetVoiceParam(self.synth.eci, 0, eloquence.eciSpeed, self.rate)
  elif data == 'p' and self.fs is not None:
   self.fs.pause()
  elif data == 'c' and self.fs is not None:
   self.fs.play(False)
  elif data == 't':
   self.push("active sockets: %d\n" % len(self.server.sockets))

 def handle_close(self):
  asynchat.async_chat.handle_close(self)
  if self in self.server.sockets:
   self.server.sockets.remove(self)

 def reset_wave(self):
  if self.wave is not None:
   self.wave.close()
  self.output_io.truncate(0)
  self.wave = wave.open(self.output_io, 'w')
  self.wave.setnchannels(1)
  self.wave.setsampwidth(2)
  self.wave.setframerate(11025)

 def _paramToPercent(cls, current, min, max):
  """Convert a raw parameter value to a percentage given the current, minimum and maximum raw values.
  @param current: The current value.
  @type current: int
  @param min: The minimum value.
  @type current: int
  @param max: The maximum value.
  @type max: int
  """
  return int(round(float(current - min) / (max - min) * 100))

 def _percentToParam(cls, percent, min, max):
  """Convert a percentage to a raw parameter value given the current percentage and the minimum and maximum raw parameter values.
		@param percent: The current percentage.
		@type percent: int
		@param min: The minimum raw parameter value.
		@type min: int
		@param max: The maximum raw parameter value.
		@type max: int
		"""
  return int(round(float(percent) / 100 * (max - min) + min))

 def stop(self):
  if self.fs is not None:
   self.fs.stop()
   self.fs.free()
   self.fs = None

 def speak_sequence(self, sequence):
  print sequence
  out = []
  for i, item in enumerate(sequence):
   if (isinstance(item, basestring)):
    if ('skp' in self.ctxt or item.strip(' "') == ''):
     continue
    item = item.strip(' "')
    item = item.replace('`', ' ')
    item = item.replace(' , ', ' `p1, ')
    out.append(item)
   if item[0] == 'ctxt':
    self.ctxt = item[1].lower()
   elif item[0] == 'slnc':
    out.append('`p%d' % int(item[1]))
    if len(sequence) != 1 and i != 0:
     out[-1]+='. '
   elif item[0] == 'pbas':
    new_pitch = self.pitch + int(float(item[1]))
    out.append('`vb%d' % new_pitch)
    self.pitch = new_pitch
  outstr = " ".join(out)
  self.synth.addText('`pp0 '+outstr)
  self.synth.synthesize()
  self.synth.synchronize()

class Dispatcher(asyncore.dispatcher):

 def __init__(self, port):
  asyncore.dispatcher.__init__(self)
  self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
  self.bind(('', port))
  self.listen(5)
  self.sockets = []

 def handle_accept(self):
  conn, addr = self.accept()
  sock = Receiver(self, conn, addr)
  self.sockets.append(sock)


s = Dispatcher(port)
if os.path.exists('ready.ogg'):
 stream = sound_lib.stream.FileStream(file='ready.ogg', autofree=True)
 stream.play(True)

while True:
 asyncore.poll(0.05)
 for rec in s.sockets:
  if rec.fs is not None and not rec.fs.is_playing and not rec.fs.is_paused:
   rec.push('x\n')
   rec.fs.free()
   rec.fs = None

